/* dirmngr.c - LDAP access
 *	Copyright (C) 2002 Klarlvdalens Datakonsult AB
 *
 * This file is part of DirMngr.
 *
 * DirMngr is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DirMngr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <assert.h>

#define JNLIB_NEED_LOG_LOGV
#include "crlcache.h"
#include "cert.h"
#include "crlfetch.h"
#include "dirmngr.h"

#include "../assuan/assuan.h" /* needed for the malloc hooks */

/* no i18n needed */
#define N_(a) a
#define _(a) a

#define MAXLEN_CERT 4096
#define GET_PTR( ctx ) ((struct server_control_s*)assuan_get_pointer(ctx))->server_local

/* Data used to associate an Assuan context with local server data */
struct server_local_s {
  ASSUAN_CONTEXT assuan_ctx;
  CrlCache*      crl_cache;
  int message_fd;
  int recursion;
};

static int inquire_cert( ASSUAN_CONTEXT ctx, char* line )
{
  char *value = NULL;
  size_t valuelen; 
  int rc;
  char* buf;
  char* issuer;
  struct server_local_s* ptr = GET_PTR( ctx );
  KsbaCert issuer_cert;

  buf = xmalloc( (strlen(line))+10 * sizeof( char ) );
  strcpy( buf, "SENDCERT");
  /*strcat( buf, line );*/
  rc = assuan_inquire( ctx, buf,
		       &value, &valuelen, MAXLEN_CERT);
  xfree( buf );
  if( rc ) return rc;
  
  issuer_cert = make_cert( value, valuelen );
  if( issuer_cert == NULL ) {
    dirmngr_debug( "Error in issuer cert\n");
    return ASSUAN_General_Error;
  }
  issuer = ksba_cert_get_issuer( issuer_cert, 0 );
  xfree( value ); value = NULL;
  ksba_cert_release( issuer_cert );
  if( !issuer ) {
    return ASSUAN_No_CRL_For_Cert; /* FIXME: return better error */    
  }
  rc = crl_fetch( issuer, &value, &valuelen );
  ksba_free( issuer );  
  if( rc ) {
    xfree( value );
    dirmngr_debug( "Error in crl_fetch\n");
    return ASSUAN_No_CRL_For_Cert;
  }
  rc = crl_cache_insert( ptr->crl_cache, value, valuelen );
  xfree( value );
  if( rc ) {
    /* ??? */
    if( rc == CRL_TOO_OLD ) return ASSUAN_CRL_Too_Old;
    else return ASSUAN_No_CRL_For_Cert;    
  }
  return rc;
}

static int
cmd_isvalid( ASSUAN_CONTEXT ctx, char* line )
{
  struct server_local_s* ptr = GET_PTR( ctx );
  char* ser = line;
  char* saveline = xstrdup( line );
  int rc = 0;

  while( *ser != '\0' && *ser != '.' ) ++ser;
  if( *ser != '.' ) {
    dirmngr_debug( "wrong format \"%s\"\n", line);  
    return ASSUAN_General_Error;
  }
    
  /* chop it up */
  *ser = '\0'; ++ser;
  dirmngr_debug( "line is  \"%s\"\n", line );

  switch( crl_cache_isvalid( ptr->crl_cache, line, ser ) ) {
  case CRL_CACHE_VALID:
    rc = ASSUAN_No_Error;
    //fprintf(stderr, "VALID\n");
    break;
  case CRL_CACHE_INVALID:
    rc = ASSUAN_Cert_Revoked;
    //fprintf(stderr, "INVALID\n");
    break;
  case CRL_CACHE_DONTKNOW:
    //fprintf(stderr, "DONTKNOW\n");
    if( (rc = inquire_cert( ctx, line )) != ASSUAN_No_Error ) {
      return rc;
    }
    /* do recursive call to get result */
    if( ptr->recursion == 0 ) {
      dirmngr_debug( "Calling recursively with \"%s\"\n", saveline );
      ptr->recursion++;
      rc = cmd_isvalid( ctx, saveline );
      ptr->recursion--;
    } else {
      /* we already tried recursively, report an error */
      rc = ASSUAN_No_CRL_For_Cert;
    }
    break;
  }
  /* restore line */
  --ser; *ser = '.';  
  xfree( saveline );
  return rc;
}

static int
cmd_lookup( ASSUAN_CONTEXT ctx, char* line )
{
  return 0;
}

/* Tell the assuan library about our commands */
static int
register_commands (ASSUAN_CONTEXT ctx)
{
  static struct {
    const char *name;
    int cmd_id;
    int (*handler)(ASSUAN_CONTEXT, char *line);
  } table[] = {
    { "ISVALID",    0,  cmd_isvalid },
    { "LOOKUP",     0,  cmd_lookup },
    { "",     ASSUAN_CMD_INPUT, NULL },
    { "",     ASSUAN_CMD_OUTPUT, NULL },
    { NULL }
  };
  int i, j, rc;

  for (i=j=0; table[i].name; i++)
    {
      rc = assuan_register_command (ctx,
                                    table[i].cmd_id? table[i].cmd_id
                                                   : (ASSUAN_CMD_USER + j++),
                                    table[i].name, table[i].handler);
      if (rc)
        return rc;
    }
  return 0;
}


/* Startup the server */
void
start_command_handler (void)
{
  int rc;
  int filedes[2];
  ASSUAN_CONTEXT ctx;
  struct server_control_s ctrl;

  memset (&ctrl, 0, sizeof ctrl);


  /* For now we use a simple pipe based server so that we can work
     from scripts.  We will later add options to run as a daemon and
     wait for requests on a Unix domain socket */
  filedes[0] = 0;
  filedes[1] = 1;
  rc = assuan_init_pipe_server (&ctx, filedes);
  if (rc)
    {
      log_error ("failed to initialize the server: %s\n",
                 assuan_strerror(rc));
      dirmgr_exit (2);
    }

  rc = register_commands (ctx);
  if (rc)
    {
      log_error ("failed to the register commands with Assuan: %s\n",
                 assuan_strerror(rc));
      dirmgr_exit (2);
    }

  ctrl.server_local = xcalloc (1, sizeof *ctrl.server_local);
  ctrl.server_local->assuan_ctx = ctx;
  ctrl.server_local->crl_cache = crl_cache_init();
  ctrl.server_local->message_fd = -1;
  ctrl.server_local->recursion = 0;

  assuan_set_pointer (ctx, &ctrl);

  for (;;) {
    rc = assuan_accept (ctx);
    if (rc == -1) {
      break;
    }
    else if (rc) {
      log_info ("Assuan accept problem: %s\n", assuan_strerror (rc));
      break;
    }

    rc = assuan_process (ctx);
    if (rc) {
      log_info ("Assuan processing failed: %s\n", assuan_strerror (rc));
      continue;
    }
  }

  crl_cache_deinit(  ctrl.server_local->crl_cache );
  assuan_deinit_server (ctx);
}
